/****************************************************************************
*                   partsys.c
*
*  This module contains functions for the particle_system feature.
*
*****************************************************************************/

#include "frame.h"

#ifdef HuffParticlePatch

#include "vector.h"
#include "povproto.h"
#include "colour.h"
#include "express.h"
#include "pigment.h"
#include "objects.h"
#include "blob.h"
#include "warps.h"
#include "partsys.h"
#include "splines.h"
#include "pattern.h"



/*****************************************************************************
* Local preprocessor defines
******************************************************************************/

#define VAssign(D,S) {(D)[X] = S[X];(D)[Y] = S[Y];(D)[Z] = S[Z];}
#define VEqEq(A,B) ( ((A)[X] == B[X]) && ((A)[Y] == B[Y]) && ((A)[Z] == B[Z]) )



/*****************************************************************************
* Local typedefs
******************************************************************************/
/*****************************************************************************
* Static functions
******************************************************************************/

static PART * Create_Part();
static void Emit_Part(PART_SYS * sys, PART_EMITTER * emitter);

static void Update_Liquid_Particle(PART_SYS * sys, PART * part);
static void Update_Atmospheric_Particle(PART_SYS * sys, PART * part);
/*static void Update_Solid_Particle(PART_SYS * sys, PART * part);*/

static int Do_Particle_Collisions(PART_SYS * sys, PART * part, VECTOR displacement);

static DBL Rand(int stream);
static DBL SRand(int stream);
static DBL RRand(int stream, DBL MinVal, DBL MaxVal);
static DBL GRand(int stream, DBL MinVal, DBL MaxVal);

/*Unsigned rand*/
static DBL Rand(int stream)
{
    return stream_rand(stream);
}
/*Signed rand*/
static DBL SRand(int stream)
{
    return stream_rand(stream)*2-1;
}
/*Ranged rand*/
static DBL RRand(int stream, DBL MinVal, DBL MaxVal)
{
    return stream_rand(stream)*(MaxVal-MinVal)+MinVal;
}
/*Gaussian rand*/
static DBL GRand(int stream, DBL MinVal, DBL MaxVal)
{
    DBL tmp = (stream_rand(stream)+stream_rand(stream)+stream_rand(stream))/3;
    return tmp*(MaxVal-MinVal)+MinVal;
}

/*****************************************************************************
* Local variables
******************************************************************************/
/*
static METHODS Part_Sys_Obj_Methods =
{
    All_Part_Sys_Obj_Intersections,
    Inside_Part_Sys_Obj,
    Part_Sys_Obj_Normal,
    Part_Sys_Obj_UVCoord,
    (COPY_METHOD)Copy_Part_Sys_Obj,
    Translate_Part_Sys_Obj, Rotate_Part_Sys_Obj, Scale_Part_Sys_Obj, Transform_Part_Sys_Obj,
    Invert_Part_Sys_Obj, Destroy_Part_Sys_Obj
};
*/

/*****************************************************************************
*
* FUNCTION   Create_Particle_System()
*
*   
*
* INPUT
*
* OUTPUT
*
* RETURNS   pointer to a PART_SYS struct
*
* AUTHOR
*
*   Chris Huff
*
* DESCRIPTION
*
*   Creates and initializes a particle system.
*
* CHANGES
*
*   June 2000: Creation.
*
******************************************************************************/
PART_SYS * Create_Particle_System()
{
    PART_SYS * New = (PART_SYS *)POV_MALLOC(sizeof(PART_SYS), "particle_system");
    
    New->references = 1;
    
    New->drag = 0.01;
    New->time = 0;
    New->tScale = 1;
    Make_Vector(New->grav, 0,-0.0981, 0);
    Make_Vector(New->wind, 0, 0, 0);
    Make_Vector(New->jitter, 0.05,0.05, 0.05);
    New->env = NULL;
    New->temperature_pigment = NULL;
    New->emitter_list = NULL;
    New->particle_list = NULL;
    New->space_warp_list = NULL;
    
    New->num_of_particles = 0;
    
    New->particles_created = 0;
    New->particles_killed = 0;
    New->particle_bond_interactions = 0;
    New->particle_drag_interactions = 0;
    New->particle_collisions = 0;
    New->num_of_emitters = 0;
    New->num_of_space_warps = 0;
    New->max_velocity = 0;
    
    New->seed = stream_seed(159753);
    
    return New;
}


/*****************************************************************************
*
* FUNCTION   Copy_Particle_System()
*
*   
*
* INPUT
*
* OUTPUT
*
* RETURNS   pointer to a PART_SYS struct
*
* AUTHOR
*
*   Chris Huff
*
* DESCRIPTION
*
*   Copies a particle system.
*
* CHANGES
*
*   June 2000: Creation.
*
******************************************************************************/
PART_SYS * Copy_Particle_System(PART_SYS * src)
{
    /*Currently don't do deep copy, share data*/
    src->references++;
    return src;
}


/*****************************************************************************
*
* FUNCTION   Destroy_Particle_System()
*
*   
*
* INPUT
*
* OUTPUT
*
* RETURNS   pointer to a PART_SYS struct
*
* AUTHOR
*
*   Chris Huff
*
* DESCRIPTION
*
*   Destroys a particle system, including all its particles, emitters, and
*   space_warps.
*
* CHANGES
*
*   June 2000: Creation.
*
******************************************************************************/
void Destroy_Particle_System(PART_SYS * sys)
{
    PART_EMITTER * em = sys->emitter_list;
    SPACE_WARP * wp = sys->space_warp_list;
    PART * part = sys->particle_list;
    unsigned long int particles_destroyed = 0;
    
    if(sys->references <= 1)
    {
        while(em != NULL)
        {
            PART_EMITTER * temp = em;
            em = em->next;
            if(temp->obj != NULL)
            {   Destroy_Object(temp->obj);}
            if(temp->trans != NULL)
            {   Destroy_Transform(temp->trans);}
    /*        if(temp->texture != NULL)
            {   Destroy_Textures(temp->texture);}*/
            POV_FREE(temp);
        }
        
        while(wp != NULL)
        {
            SPACE_WARP * temp = wp;
            wp = wp->next;
            if(temp->trans != NULL)
            {   Destroy_Transform(temp->trans);}
            if(temp->warp != NULL)
            {   Destroy_TPat_Fields(temp->warp);
                POV_FREE(temp->warp);}
            POV_FREE(temp);
        }
        
        while(part != NULL)
        {
            PART * temp = part;
            part = part->next;
            POV_FREE(temp);
            particles_destroyed++;
            COOPERATE_0
            COOPERATE_1
        }
        Debug_Info("%d particles destroyed\n", particles_destroyed);
        
        if(sys->temperature_pigment != NULL)
            Destroy_Pigment(sys->temperature_pigment);
        
        if(sys->env != NULL)
            Destroy_Object(sys->env);
        
        {
            int s = sizeof(PART);
            Debug_Info("sizeof(PART) = %d\n", s);
        }
    }
    else
    {
        sys->references--;
    }
}


void Destroy_Particle_Emitter_List(PART_SYS * sys)
{
    PART_EMITTER * em = sys->emitter_list;
    
    while(em != NULL)
    {
        PART_EMITTER * temp = em;
        em = em->next;
        if(temp->obj != NULL)
        {   Destroy_Object(temp->obj);}
        if(temp->trans != NULL)
        {   Destroy_Transform(temp->trans);}
/*        if(temp->texture != NULL)
        {   Destroy_Textures(temp->texture);}*/
        POV_FREE(temp);
    }
}


void Destroy_Space_Warp_List(PART_SYS * sys)
{
    SPACE_WARP * wp = sys->space_warp_list;
    
    while(wp != NULL)
    {
        SPACE_WARP * temp = wp;
        wp = wp->next;
        if(temp->trans != NULL)
        {   Destroy_Transform(temp->trans);}
        if(temp->warp != NULL)
        {   Destroy_TPat_Fields(temp->warp);
            POV_FREE(temp->warp);}
        POV_FREE(temp);
    }
}


/*****************************************************************************
*
* FUNCTION   Create_Particle_Emitter()
*
*   
*
* INPUT
*
* OUTPUT
*
* RETURNS   pointer to a PART_EMITTER struct
*
* AUTHOR
*
*   Chris Huff
*
* DESCRIPTION
*
*   .
*
* CHANGES
*
*   June 2000: Creation.
*
******************************************************************************/
PART_EMITTER * Create_Particle_Emitter()
{
    PART_EMITTER * New = (PART_EMITTER *)POV_MALLOC(sizeof(PART_EMITTER), "particle_emitter");
    
    New->type = FLUID_PARTICLE;
    New->radius = 1;
    New->strength = 1;
    New->elasticity = 0.5;
    New->age = 0;
    New->lifespan = -1;
    New->lifespan_jitter = 0;
    New->mass = 1;
    New->mass_jitter = 0;
    New->temperature = 0;
    Make_Vector(New->pos, 0, 0, 0);
    Make_Vector(New->vel, 0, 1, 0);
    
    New->ipf_strength = 0;
    New->ipf_exp = 1;
    New->ipf_dist = -1;
    New->ipf_search_rad = -1;
    New->ipf_drag = 0;
    New->ipf_type = NO_IPF;
    New->ipf_spline = NULL;
    
    New->start_time = 0;
    New->stop_time = -1;
    New->particle_stretch = 1;
    New->min_angle = 0;
    New->max_angle = 1;
    New->emission_rate = 1;
    New->last_emitted = 0;
    New->emitter_type = POINT_EMITTER;
    New->obj = NULL;
    New->texture = NULL;
    New->max_tests = 100;
    New->trans = Create_Transform();
    New->next = NULL;
    
    return New;
}


/*****************************************************************************
*
* FUNCTION   Add_Particle_Emitter()
*
*   
*
* INPUT    PART_SYS *, PART_EMITTER *
*
* OUTPUT
*
* RETURNS
*
* AUTHOR
*
*   Chris Huff
*
* DESCRIPTION
*
*   Adds a particle emitter into a particle system
*
* CHANGES
*
*   June 2000: Creation.
*
******************************************************************************/
void Add_Particle_Emitter(PART_SYS * sys, PART_EMITTER * em)
{
    if(sys->emitter_list == NULL)
    {   sys->emitter_list = em;}
    else
    {   em->next = sys->emitter_list;
        sys->emitter_list = em;}
    sys->num_of_emitters++;
}


/*****************************************************************************
*
* FUNCTION   Create_Space_Warp()
*
*   
*
* INPUT
*
* OUTPUT
*
* RETURNS   pointer to a SPACE_WARP struct
*
* AUTHOR
*
*   Chris Huff
*
* DESCRIPTION
*
*   .
*
* CHANGES
*
*   June 2000: Creation.
*
******************************************************************************/
SPACE_WARP * Create_Space_Warp()
{
    SPACE_WARP * New = (SPACE_WARP *)POV_MALLOC(sizeof(SPACE_WARP), "space_warp");
    
    New->type = GRAVITY_SOURCE;
    New->radius = 1;
    New->strength = 1;
    New->falloff = 2;
    Make_Vector(New->pos, 0, 0, 0);
    Make_Vector(New->vect_dat_a, 0, 1, 0);
    
    New->warp = (TPATTERN *)POV_MALLOC(sizeof (TPATTERN), "TPATTERN");
    Init_TPat_Fields(New->warp);
    New->warp->Type = BOZO_PATTERN;
    
    New->trans = Create_Transform();
    New->next = NULL;
    
    return New;
}

 
/*****************************************************************************
*
* FUNCTION   Add_Space_Warp()
*
*   
*
* INPUT    PART_SYS *, SPACE_WARP *
*
* OUTPUT
*
* RETURNS
*
* AUTHOR
*
*   Chris Huff
*
* DESCRIPTION
*
*   Adds a space warp into a particle system.
*
* CHANGES
*
*   June 2000: Creation.
*
******************************************************************************/
void Add_Space_Warp(PART_SYS * sys, SPACE_WARP * wp)
{
    if(sys->space_warp_list == NULL)
    {   sys->space_warp_list = wp;}
    else
    {   wp->next = sys->space_warp_list;
        sys->space_warp_list = wp;}
    sys->num_of_space_warps++;
}


/*****************************************************************************
*
* FUNCTION   Create_Part()
*
*   
*
* INPUT    
*
* OUTPUT
*
* RETURNS
*
* AUTHOR
*
*   Chris Huff
*
* DESCRIPTION
*
*   Creates a particle.
*
* CHANGES
*
*   June 2000: Creation.
*
******************************************************************************/
static PART * Create_Part()
{
    PART * New = (PART *)POV_MALLOC(sizeof(PART), "particle");
    New->age = 0;
    New->temperature = 0;
    Make_Vector(New->oldPos, 0, 0, 0);
    Make_Vector(New->oldVel, 0, 0, 0);
    Make_Vector(New->pos, 0, 0, 0);
    Make_Vector(New->vel, 0, 0, 0);
    New->parent_emitter = NULL;
    New->prev = NULL;
    New->next = NULL;
    
    return New;
}


/*****************************************************************************
*
* FUNCTION   Emit_Part()
*
*   
*
* INPUT    pointer to a PART_EMITTER and a PART_SYS
*
* OUTPUT
*
* RETURNS
*
* AUTHOR
*
*   Chris Huff
*
* DESCRIPTION
*
*   Emits a particle from an emitter.
*
* CHANGES
*
*   June 2000: Creation.
*
******************************************************************************/
static void Emit_Part(PART_SYS * sys, PART_EMITTER * emitter)
{
    VECTOR tempVect;
    VECTOR tempVect2;
    DBL tempDbl;
    TRANSFORM Trans;
    DBL speed = 0;
    PART * New = Create_Part();
/*    PART * New = (PART *)POV_MALLOC(sizeof(PART), "particle");*/
    
    VLength(speed, emitter->vel);
    
    New->age = emitter->age;
    New->temperature = emitter->temperature;
    New->prev = NULL;
    New->next = NULL;
    
    New->parent_emitter = emitter;
    
    
    /*Calculate particle start position*/
    switch(emitter->emitter_type)
    {
        case POINT_EMITTER:
        {
            Assign_Vector(New->pos, emitter->pos);
        }
        break;
        case OBJECT_EMITTER:
        {
            if(emitter->obj != NULL)
            {
                VECTOR minExt;
                VECTOR maxExt;
                VECTOR pt;
                int k=0;
                Make_min_max_from_BBox(minExt, maxExt, emitter->obj->BBox);
                
                do
                {
                    VECTOR jitterVect;
                    Make_Vector(jitterVect, SRand(sys->seed), SRand(sys->seed), SRand(sys->seed));
                    
                    Assign_Vector(pt, maxExt); VSubEq(pt, minExt);
                    VEvaluateEq(pt, jitterVect); VAddEq(pt, minExt);
                    k++;
                }
                while((!Inside_Object(pt, emitter->obj)) && k<emitter->max_tests);
                
                Assign_Vector(New->pos, pt);
            }
        }
        break;
        default:
            Assign_Vector(New->pos, emitter->pos);
    }
    MTransPoint(New->pos, New->pos, emitter->trans);
    Assign_Vector(New->oldPos, New->pos);
    
    
    /*Calculate particle start velocity*/
    Assign_Vector(New->vel, emitter->vel);
    if((New->vel[X] != 0)||(New->vel[Y] != 0)||(New->vel[Z] != 0))
    {
        /*rotate from axis*/
        if(emitter->max_angle > 0)
        {   if(emitter->min_angle > 0)
            {   tempDbl = RRand(sys->seed, emitter->min_angle, emitter->max_angle);}
    	    else
    	    {   tempDbl = emitter->max_angle;}}
    	else
        {   tempDbl = Rand(sys->seed)*emitter->min_angle;}
    	Make_Vector(tempVect2, 0.6, 0.3, 1);/*Pick a vector not likely to be used as direction*/
        VCross(tempVect, tempVect2, emitter->vel);
        VNormalizeEq(tempVect);
        VScaleEq(tempVect, tempDbl);
        
        Compute_Rotation_Transform(&Trans, tempVect);
        MTransPoint(New->vel, emitter->vel, &Trans);
        
        /*rotate around axis*/
    	tempDbl = Rand(sys->seed)*360.0;
        Compute_Axis_Rotation_Transform(&Trans,emitter->vel,tempDbl);
        MTransPoint(New->vel, New->vel, &Trans);
            
        /*transform velocity*/
        MTransPoint(New->vel, New->vel, emitter->trans);
        VNormalizeEq(New->vel); VScaleEq(New->vel, speed);
    }
    
    Assign_Vector(New->oldVel, New->vel);
    
    
    if(sys->particle_list == NULL)
    {
        sys->particle_list = New;
    }
    else
    {
        sys->particle_list->prev = New;
        New->next = sys->particle_list;
        sys->particle_list = New;
    }
}


/*****************************************************************************
*
* FUNCTION   Update_Liquid_Particle()
*            Update_Atmospheric_Particle()
*            Update_Solid_Particle()
*
*   
*
* INPUT    PART_SYS *, PART *
*
* OUTPUT
*
* RETURNS
*
* AUTHOR
*
*   Chris Huff
*
* DESCRIPTION
*
*   Updates particle position and velocity.
*
* CHANGES
*
*   July 2000: Creation.
*
******************************************************************************/
static void Calculate_IPF(PART_SYS * sys, PART * part, VECTOR force)
{
    PART * part2;
    VECTOR local_drag;
    unsigned int drag_samples = 0;
    Make_Vector(local_drag,0,0,0);
    switch(part->parent_emitter->ipf_type)
    {
        case NO_IPF:
            return;
        break;
        case IPF_A:
            /*Do inter-particle bond forces...*/
            part2 = sys->particle_list;
            while((part2 != NULL)&&(part->parent_emitter->ipf_dist > 0)&&(part->parent_emitter->ipf_strength != 0))
            {
                COOPERATE_0
                COOPERATE_1
                if(part2 != part)
                {/*don't check a particle against itself*/
                    DBL dist;
                    VECTOR dir;
                    VDist(dist, part->pos, part2->oldPos);
                    if(dist<=part->parent_emitter->ipf_dist)
                    {
                        Assign_Vector(dir, part->pos);
                        VSubEq(dir, part2->oldPos);
                        /*strength*(1.5x-0.75)/((x-1.5)(x+0.5))*/
                        dist /= part->parent_emitter->ipf_dist;
                        /*Get_Spline_Val(spline,Val,Express);*/
                        dist = part->parent_emitter->ipf_strength*(1.5*dist - 0.75)/
                                ((dist-1.5)*(dist+0.5));
                        VNormalizeEq(dir);
                        VScaleEq(dir, dist);
                        VAddEq(force, dir);
                        sys->particle_bond_interactions++;
                    }
                }
                part2 = part2->next;
            }
        break;
        case CUSTOM_IPF:
            /*Do inter-particle bond forces...*/
            part2 = sys->particle_list;
            while((part2 != NULL)&&(part->parent_emitter->ipf_dist > 0))
            {
                COOPERATE_0
                COOPERATE_1
                if(part2 != part)
                {/*don't check a particle against itself*/
                    DBL dist;
                    VECTOR dir;
                    VDist(dist, part->pos, part2->oldPos);
                    if(dist<=part->parent_emitter->ipf_dist)
                    {
                        EXPRESS tmpExpr;
                        Assign_Vector(dir, part->pos);
                        VSubEq(dir, part2->oldPos);
                        /*strength*(1.5x-0.75)/((x-1.5)(x+0.5))*/
                        dist /= part->parent_emitter->ipf_dist;
                        /*Get_Spline_Val(spline,Val,Express);*/
                        Get_Spline_Val(part->parent_emitter->ipf_spline,dist,tmpExpr);
                        dist = tmpExpr[0];
                        VNormalizeEq(dir);
                        VScaleEq(dir, dist);
                        VAddEq(force, dir);
                        sys->particle_bond_interactions++;
                    }
                }
                part2 = part2->next;
            }
        break;
    }
    
    /*Do inter-particle drag forces...*/
    if((part->parent_emitter->ipf_drag != 0)&&(part->parent_emitter->ipf_search_rad > 0))
    {
        part2 = sys->particle_list;
        while(part2 != NULL)
        {
            COOPERATE_0
            COOPERATE_1
            if(part2 != part)
            {/*don't check a particle against itself*/
                DBL dist;
                VDist(dist, part->pos, part2->oldPos);
                if(dist<=part->parent_emitter->ipf_search_rad)
                {
                    VAddEq(local_drag, part2->oldVel);
                    drag_samples++;
                    sys->particle_drag_interactions++;
                }
            }
            part2 = part2->next;
        }
        if(drag_samples > 0)
        {
            VInverseScaleEq(local_drag, drag_samples);
            VSubEq(local_drag, part->vel);
            VScaleEq(local_drag, part->parent_emitter->ipf_drag);
            VAddEq(force, local_drag);
        }
    }
}
/*****************************************************************************
*
* FUNCTION   Update_Liquid_Particle()
*            Update_Atmospheric_Particle()
*            Update_Solid_Particle()
*
*   
*
* INPUT    PART_SYS *, PART *
*
* OUTPUT
*
* RETURNS
*
* AUTHOR
*
*   Chris Huff
*
* DESCRIPTION
*
*   Updates particle position and velocity.
*
* CHANGES
*
*   July 2000: Creation.
*
******************************************************************************/
static void Update_Liquid_Particle(PART_SYS * sys, PART * part)
{
    DBL speed;
    VECTOR force;
    VECTOR posOffset;
    VECTOR temp;
    VECTOR jitterVect;
    SPACE_WARP * warp;
    int collision = FALSE;
    Make_Vector(force,0,0,0);
    VLength(speed, part->vel);
    
    sys->max_velocity = max(sys->max_velocity, speed);
    
    /*calculate forces on particle*/
    VAssign(temp, sys->wind); VSubEq(temp, part->vel); VScaleEq(temp, sys->drag);
    VAddEq(force, temp);
    
    VAddEq(force, sys->grav);
    
    Calculate_IPF(sys, part, force);
    /*Do inter-particle bond forces...*/
    
    /*Add in effect of space_warps*/
    warp = sys->space_warp_list;
    while(warp != NULL)
    {
        VECTOR temp;
        VECTOR temp2;
        Assign_Vector(temp, part->pos);
        MInvTransPoint(temp, temp, warp->trans);
        Assign_Vector(temp2, temp);
        Warp_EPoint(temp2, part->pos, warp->warp, FALSE);
        VSubEq(temp, temp2);
        VAddEq(force, temp);
        warp = warp->next;
    }
    
    COOPERATE_0
    COOPERATE_1
    
    /*Calculate offset*/
    VAssign(posOffset, part->vel); VScaleEq(posOffset, sys->tScale);
/*    VAssign(temp, force); VInverseScaleEq(temp, part->parent_emitter->mass); VScaleEq(temp, sys->tScale); VInverseScaleEq(temp, 2);
    VAddEq(posOffset, temp);*//*Removed effects of force on particle movement, thanks Christophe Bouffartigue*/
    
    /*check for collision*/
    collision = Do_Particle_Collisions(sys, part, posOffset);
    COOPERATE_0
    COOPERATE_1
	
	/*Move particle, update velocity*/
	if(collision == FALSE)
	{/*No collision, move particle as usual*/
	    VAddEq(part->pos, posOffset);
    }
	
	VAssign(temp, force); VScaleEq(temp, sys->tScale); VInverseScaleEq(temp, part->parent_emitter->mass);
	VAddEq(part->vel, temp);
	
	/*Jitter velocity*/
    Make_Vector(jitterVect, SRand(sys->seed), SRand(sys->seed), SRand(sys->seed));
    VEvaluateEq(jitterVect, sys->jitter);
    VAddEq(part->vel, jitterVect);
    
    part->age += sys->tScale;
}

static void Update_Atmospheric_Particle(PART_SYS * sys, PART * part)
{
    DBL speed;
    VECTOR force;
    VECTOR posOffset;
    VECTOR temp;
    VECTOR buoyancy;
    VECTOR jitterVect;
    SPACE_WARP * warp;
    COLOUR tempCol;
    DBL relTemp;
    int collision = FALSE;
    Make_Vector(force,0,0,0);
    VLength(speed, part->vel);
    
    sys->max_velocity = max(sys->max_velocity, speed);
    
    /*calculate forces on particle*/
    VAssign(temp, sys->wind); VSubEq(temp, part->vel); VScaleEq(temp, sys->drag);
    VAddEq(force, temp);
    
    VAddEq(force, sys->grav);
    
    
    if(sys->temperature_pigment != NULL)
    {   Compute_Pigment(tempCol, sys->temperature_pigment, part->pos, NULL);
        relTemp = GREY_SCALE(tempCol) - part->temperature;}
    else
    {   relTemp = -part->temperature;}
    
    VAssign(buoyancy, sys->grav);
    VScaleEq(buoyancy, relTemp);
    VAddEq(force, buoyancy);
    
    /*Do inter-particle drag forces...*/
    Calculate_IPF(sys, part, force);
    
    /*Add in effect of space_warps*/
    warp = sys->space_warp_list;
    while(warp != NULL)
    {
        VECTOR temp;
        VECTOR temp2;
        Assign_Vector(temp, part->pos);
        MInvTransPoint(temp, temp, warp->trans);
        Assign_Vector(temp2, temp);
        Warp_EPoint(temp2, part->pos, warp->warp, FALSE);
        VSubEq(temp, temp2);
        VAddEq(force, temp);
        warp = warp->next;
    }
    
    COOPERATE_0
    COOPERATE_1
    
    /*Calculate offset*/
    VAssign(posOffset, part->vel); VScaleEq(posOffset, sys->tScale);
    VAssign(temp, force); VInverseScaleEq(temp, part->parent_emitter->mass); VScaleEq(temp, sys->tScale); VInverseScaleEq(temp, 2);
    VAddEq(posOffset, temp);
    
    /*check for collision*/
    collision = Do_Particle_Collisions(sys, part, posOffset);
    COOPERATE_0
    COOPERATE_1
	
	/*Move particle, update velocity*/
	if(collision == FALSE)
	{/*No collision, move particle as usual*/
	    VAddEq(part->pos, posOffset);
    }
	
	VAssign(temp, force); VScaleEq(temp, sys->tScale); VInverseScaleEq(temp, part->parent_emitter->mass);
	VAddEq(part->vel, temp);
	
	/*Jitter velocity*/
    Make_Vector(jitterVect, SRand(sys->seed), SRand(sys->seed), SRand(sys->seed));
    VEvaluateEq(jitterVect, sys->jitter);
    VAddEq(part->vel, jitterVect);
    
    part->age += sys->tScale;
    
    part->temperature += relTemp/part->parent_emitter->mass;
}
/*****************************************************************************
*
* FUNCTION   Do_Particle_Collisions()
*
*   
*
* INPUT    PART_SYS *, PART *
*
* OUTPUT
*
* RETURNS  TRUE if collision encountered
*
* AUTHOR
*
*   Chris Huff
*
* DESCRIPTION
*
*   Checks for particle collisions and updates particle.
*
* CHANGES
*
*   July 2000: Creation.
*
******************************************************************************/
static int Do_Particle_Collisions(PART_SYS * sys, PART * part, VECTOR displacement)
{/*Note: modify to recursively check for collision immediately following a collision*/
    int collision = FALSE;
    /*check for collision*/
    if(sys->env != NULL)
    {
        INTERSECTION Intersect;
        RAY Ray;
        VECTOR Norm;
        VECTOR Dir;
        VECTOR IPt;
        DBL dist;
        DBL speed;
        DBL colDist;
        
        Initialize_Ray_Containers(&Ray);
        
        VAssign(Dir, part->vel); VNormalizeEq(Dir);
        VAssign(Ray.Direction, Dir);
        VAssign(Ray.Initial, part->pos);
        
        if(Intersection(&Intersect, sys->env, &Ray))
        {
            Assign_Vector(IPt, Intersect.IPoint);
            VDist(dist, IPt, part->pos);
            VLength(colDist, displacement);
            if(dist < colDist)
            {
                if(part->parent_emitter->elasticity > 0)
                {
                    DBL n;
                    VECTOR posDisp;
                    collision = TRUE;
                    sys->particle_collisions++;
                    Normal(Norm, Intersect.Object, &Intersect);
                    VNormalizeEq(Norm);
                    VLength(speed, part->vel);
                    
                    /*Change velocity direction*/
                    VDot(n, Dir, Norm);
                    n *= -2.0;
                    VAddScaledEq(Dir, n, Norm);
                    VNormalizeEq(Dir);
                    
                    /*Calculate new position*/
                        Assign_Vector(part->pos, IPt);
                        Assign_Vector(posDisp, Dir); VScaleEq(posDisp, colDist-dist); VScaleEq(posDisp, part->parent_emitter->elasticity);
    	                VAddEq(part->pos, posDisp);
                    
                    /*Set new velocity*/
                    VScaleEq(Dir, speed);
                    VScaleEq(Dir, part->parent_emitter->elasticity);
                    VAssign(part->vel, Dir);
                }
                else
                {
                    Assign_Vector(part->pos, IPt);
                    Make_Vector(part->vel, 0, 0, 0);
                }
            }
        }
    }
    COOPERATE_0
    COOPERATE_1
    return collision;
}


/*****************************************************************************
*
* FUNCTION   Run_Sys()
*
*   
*
* INPUT    PART_SYS *, unsigned long int cycles, DBL time
*
* OUTPUT
*
* RETURNS
*
* AUTHOR
*
*   Chris Huff
*
* DESCRIPTION
*
*   Runs the simulation "sys" for "sim_length" amount of time, divided into
*   "cycles" iterations.
*
* CHANGES
*
*   June 2000: Creation.
*
******************************************************************************/
void Run_Sys(PART_SYS * sys, unsigned long int cycles, DBL sim_length)
{
    PART * part;
    PART * part2;
    PART_EMITTER * em;
    unsigned long int j=0;
    sys->tScale = sim_length/(DBL)cycles;
    
    Status_Info("\nRunning particle simulation...\n");

    for(j=0; j<cycles; j++)
    {
        COOPERATE_0
        COOPERATE_1
        /*update emitters and emit new particles*/
        em = sys->emitter_list;
        while(em != NULL)
        {
            if(((sys->time-em->last_emitted) > 1/em->emission_rate)&&
               (sys->time>em->start_time)&&(sys->time<em->stop_time || em->stop_time<0))
            {
                int k;
                for(k=0; k < em->emission_rate*(sys->time-em->last_emitted); k++)
                {
                    Emit_Part(sys, em);
                    sys->particles_created++;
                    sys->num_of_particles++;
                }
                em->last_emitted = sys->time;
            }
            
            em = em->next;
        }
        
        
        /*update particles*/
        part = sys->particle_list;
        while(part != NULL)
        {
            COOPERATE_0
            COOPERATE_1
            if((part->age >= (part->parent_emitter->lifespan+
                part->parent_emitter->lifespan_jitter*SRand(sys->seed)))&&
               (part->parent_emitter->lifespan > 0))
            {/*Euthanize particle. :-)*/
                PART * temp = part->next;
                
                if(part->next != NULL)
                {(part->next)->prev = part->prev;}
                
                if(part->prev != NULL)
                {(part->prev)->next = part->next;}
                
                POV_FREE(part);
                part = temp;
                sys->particles_killed++;
                sys->num_of_particles--;
            }
            else
            {
                switch(part->parent_emitter->type)
                {
                    case ATMOSPHERE_PARTICLE:
                    {
                        Update_Atmospheric_Particle(sys, part);
/*                        COLOUR tempCol;
                        VECTOR buoyancy;
                        DBL relTemp;
                        if(sys->temperature_pigment != NULL)
                        {
                            Compute_Pigment(tempCol, sys->temperature_pigment, part->pos, NULL);
                            relTemp = GREY_SCALE(tempCol) - part->temperature;
                        }
                        else
                        {
                            relTemp = -part->temperature;
                        }
                        
                        VAssign(buoyancy, sys->grav);
                        VScaleEq(buoyancy, relTemp);
                        VAddEq(force, buoyancy);
                        
                        part->temperature += relTemp/part->parent_emitter->mass;*/
                    }
                    break;
                    case FLUID_PARTICLE:
                        Update_Liquid_Particle(sys, part);
                    break;
                }
                
                part = part->next;
            }
        }

        /*Reset particle positions...*/
        part2 = sys->particle_list;
        while(part2 != NULL)
        {
            Assign_Vector(part2->oldPos, part2->pos);
            Assign_Vector(part2->oldVel, part2->vel);
            part2 = part2->next;
        }
        
        /*Update particle system time*/
        sys->time += sys->tScale;
        Status_Info("Simulation %d percent done\r", (int)j*100/(cycles-1));
    }
    
    Output_Particle_System_Statistics(sys);
    Status_Info("\nParticle simulation done.\n");
}


/*****************************************************************************
*
* FUNCTION   Calc_Field_Strength()
*
*   
*
* INPUT    PART_SYS *, VECTOR EPoint
*
* OUTPUT
*
* RETURNS
*
* AUTHOR
*
*   Chris Huff
*
* DESCRIPTION
*
*   Calculates the field value at a given point
*
* CHANGES
*
*   June 2000: Creation.
*
******************************************************************************/
DBL Calc_Field_Strength(PART_SYS * sys, VECTOR EPoint)
{
	DBL totalVal = 0;
	DBL dist = 0;
	PART * part = sys->particle_list;
	
	while(part != NULL)
	{
        COOPERATE_0
        COOPERATE_1
		VDist(dist, EPoint, part->pos);
    	if(dist < part->parent_emitter->radius)
    	{
    		totalVal += Sqr(1 - Sqr(dist/part->parent_emitter->radius));
    	}
	    
		part = part->next;
	}
	return totalVal;
}


/*****************************************************************************
*
* FUNCTION   Get_Particle()
*
*   
*
* INPUT    PART_SYS *, unsigned long int
*
* OUTPUT
*
* RETURNS PART *
*
* AUTHOR
*
*   Chris Huff
*
* DESCRIPTION
*
*   Gets a pointer to a particle by index.
*
* CHANGES
*
*   July 2000: Creation.
*
******************************************************************************/

PART * Get_Particle(PART_SYS * sys, unsigned long int idx)
{
    PART * part = sys->particle_list;
    unsigned long int k = 0;
    
    if(sys->particle_list == NULL || idx >= sys->num_of_particles)
    {   return NULL;}
    
	while((part != NULL)&&(k < idx))
	{   part = part->next; k++;}
	return part;
}


/*****************************************************************************
*
* FUNCTION   Output_Particle_System_Statistics()
*
*   
*
* INPUT    PART_SYS *, VECTOR EPoint
*
* OUTPUT
*
* RETURNS
*
* AUTHOR
*
*   Chris Huff
*
* DESCRIPTION
*
*   Outputs statistics for particle simulation.
*
* CHANGES
*
*   July 2000: Creation.
*
******************************************************************************/
void Output_Particle_System_Statistics(PART_SYS * sys)
{
    Status_Info("%d emitters\n", sys->num_of_emitters);
    Status_Info("%d space warps\n", sys->num_of_space_warps);
    Status_Info("%d particles created\n", sys->particles_created);
    Status_Info("%d particles killed\n", sys->particles_killed);
    Status_Info("%d particle bond interactions\n", sys->particle_bond_interactions);
    Status_Info("%d particle drag interactions\n", sys->particle_drag_interactions);
    Status_Info("%d particle collisions\n", sys->particle_collisions);
    Status_Info("max velocity: %f\n", sys->max_velocity);
}


/*****************************************************************************
*
* FUNCTION   
*
*   
*
* INPUT    
*
* OUTPUT
*
* RETURNS
*
* AUTHOR
*
*   Chris Huff
*
* DESCRIPTION
*
*   These functions create, copy, and destroy particle_system objects.
*
* CHANGES
*
*   June 2000: Creation.
*
******************************************************************************/

PART_SYS_OBJ * Create_Part_Sys_Obj()
{
    PART_SYS_OBJ * New = (PART_SYS_OBJ *)POV_MALLOC(sizeof(PART_SYS_OBJ), "particle_system");

    Init_Particle_Blob((BLOB *)New);
    New->sys = Create_Particle_System();
    
    return New;
}

PART_SYS_OBJ * Copy_Part_Sys_Obj(OBJECT * Object)
{
    int i;
    PART_SYS_OBJ * New = Create_Part_Sys_Obj();
    PART_SYS_OBJ * Old = (PART_SYS_OBJ *)Object;
    *New = *((PART_SYS_OBJ *)Object);
    
    New->sys = Copy_Particle_System(Old->sys);
    
    
    New->Trans = Copy_Transform(New->Trans);

    New->Data->References++;

    New->Element_Texture = (TEXTURE **)POV_MALLOC(New->Data->Number_Of_Components*sizeof(TEXTURE *), "blob texture list");

    for (i = 0; i < New->Data->Number_Of_Components; i++)
    {   New->Element_Texture[i] = Copy_Textures(Old->Element_Texture[i]);}
    return New;
}

void Destroy_Part_Sys_Obj(OBJECT * Object)
{
    PART_SYS_OBJ * Obj = (PART_SYS_OBJ *)Object;
    Destroy_Particle_System(Obj->sys);
    Destroy_Blob(Object);
}




#endif
